// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#pragma once

#include "processor.h"

namespace spex {

class Darkcal : public Processor
{
public:
    Darkcal(const gadfit::MPIVars& mpi, const SPEXSettings& settings, CKD& ckd)
      : Processor { mpi, settings, ProcLevel::dark, ckd }
    {}
    ~Darkcal() = default;

protected:
    auto processBatch(const int /*i_batch*/) -> void override;

private:
    // Diagnostic CKD
    std::vector<double> dark_chi2 {}; // Fit residual chi squared for dark fit

    // Detailed output
    size_t det_nl1a {}; // Copy of nl1a in batch or nl1a_total, but resillient against possible structure change in the algorithms
    std::vector<double> medians {}; // Also used as working variable, but set as member because of detailed output option
    std::vector<int> det_inclusion {}; // 1: In initial set. >1: Iteratively included and survived filter (in order 2,n). -1: Triggered filter, thus excluded. 0: Higher than measurement that triggered filter, so not even considered
    std::vector<double> det_medianfit_chi2 {};
};

} // namespace spex
