// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1a_calibration.h"

#include "l1a_metadata.h"
#include "spex_settings.h"

namespace spex {

auto L1ACalibration::getNumberOfImages(
  const SPEXSettings& /*settings*/,
  const std::shared_ptr<L1AMetadata>& /*meta*/,
  const int n_files) const -> int
{
    // Each file is one image
    return n_files;
}

auto L1ACalibration::organize(const SPEXSettings& settings,
                              const ProcLevel proc_level,
                              std::vector<std::shared_ptr<L1AMetadata>>& meta)
  -> void
{
    // Each file is one L1A image
    for (int i_l1a {}; i_l1a < static_cast<int>(l1a_products.size()); ++i_l1a) {
        // If some frames need to be discarded for the current
        // process, adjust the number of images (frames) that will be
        // read from the L1A file.
        meta.at(i_l1a)->number_of_images -=
          settings.discard_frames_front.at(static_cast<int>(proc_level))
          + settings.discard_frames_back.at(static_cast<int>(proc_level));
        l1a_products.at(i_l1a).meta = meta.at(i_l1a);
        l1a_products.at(i_l1a).image_offset =
          settings.discard_frames_front.at(static_cast<int>(proc_level));
    }
}

} // namespace spex
