// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "l1a_flight.h"

#include "l1a.h"
#include "l1a_metadata.h"
#include "spex_settings.h"

namespace spex {

auto L1AFlight::getNumberOfImages(const SPEXSettings& settings,
                                  const std::shared_ptr<L1AMetadata>& meta,
                                  const int n_files) const -> int
{
    if (n_files == 0) {
        return 0;
    }
    if (n_files != 1) {
        throw std::runtime_error { "number of L1A files should be exactly 1 in "
                                   "flight mode" };
    }
    return settings.l1b.image_end.value_or(meta->number_of_images - 1)
           - settings.l1b.image_start + 1;
}

auto L1AFlight::organize(const SPEXSettings& settings,
                         const ProcLevel /*proc_level*/,
                         std::vector<std::shared_ptr<L1AMetadata>>& meta)
  -> void
{
    // There is only one file so only one NetCDF object
    for (auto& l1a : l1a_products) {
        l1a.meta = meta.front();
    }
    for (int i {}; i < static_cast<int>(l1a_products.size()); ++i) {
        l1a_products.at(i).image_offset = i + settings.l1b.image_start;
        // All images cover only one image by definition
        meta.front()->number_of_images = 1;
    }
}

} // namespace spex
