// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Provides a wrapper of netCDF::NcFile that includes some convenience
// member functions and functionality relevant to SPEXone.

#pragma once

#include <netcdf>
#include <string>

namespace spex {

class SPEXSettings;

class NcObject : public netCDF::NcFile
{
private:
    // To keep things simple, the file mode can be set only once
    const netCDF::NcFile::FileMode fmode {};
    std::string filename {};

public:
    NcObject(const netCDF::NcFile::FileMode fmode,
             const std::string& filename = {});
    // Call netCDF::NcFile::open using file mode that was set in the
    // constructor. If writing or replacing a file then first always
    // write the global NetCDF attributes of the SPEXone mission.
    auto open(const std::string& filename) -> void;
    // Whether the NetCDF object has been opened. It may have been
    // closed already but at least the filename has been saved with a
    // previous call to open.
    auto hasBeenOpened() const -> bool;
    // Re-open with filename defined by an earlier open command. In
    // order to open a different file first call close and then open
    // with a different name.
    auto reOpen() -> void;
    // Write the contents of the input YAML file into the variable
    // processor_configuration
    auto writeReducedConfig(const SPEXSettings& settings) -> void;
    ~NcObject();
};

} // namespace spex
