// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#pragma once

#include "cubic_spline.h"
#include "processor.h"

namespace spex {

class PolGen : public Processor
{
public:
    PolGen(const gadfit::MPIVars& mpi, const SPEXSettings& settings, CKD& ckd)
      : Processor { mpi, settings, ProcLevel::pol, ckd }
    {}
    ~PolGen() = default;

protected:
    auto processInit() -> void override;
    // Distribute L1A products into batches by viewports
    auto prepareBatches() -> void override;
    // Process L1A images by viewport
    auto processBatch(const int i_vp) -> void;

private:
    // Cubic spline for retardance prior
    std::unique_ptr<CubicSpline> retardance_prior {};
    // Cubic splines for telescope polarization prior (FOV dependent)
    std::vector<std::vector<std::unique_ptr<CubicSpline>>> telescope_prior {};
};

} // namespace spex
