// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Class for retrieving radiometrically calibrated S+ and S- spectra
// for one field of view

#pragma once

#include "ckd.h"
#include "constants.h"
#include "shared_array.h"

#include <vector>

namespace spex {

class Spectra
{
public:
    std::array<std::vector<double>, dims::pol> signal {};
    std::array<std::vector<double>, dims::pol> noise {};
    // Spectral mask
    std::array<std::vector<bool>, dims::pol> mask {};
    Spectra() = default;
    // Extract radiometrically uncalibrated S+ and S- spectra and the
    // corresponding noise spectra for a field of view
    auto extractFOV(const CKD& ckd,
                    const SharedArray<double>& image,
                    const SharedArray<double>& image_noise,
                    const SharedArray<bool>& pixel_mask,
                    const int i_fov) -> void;
    // Radiometrically calibrate
    auto calibrate(const CKD& ckd, const double exposure_time, const int i_fov)
      -> void;
    // Interpolate S+/S- to a common wavelength grid
    // (ckd.wave_common). There is no mask after this.
    auto interpolateToCommon(const CKD& ckd, const int i_fov) -> void;
};

} // namespace spex
