// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#pragma once

#include "processor.h"
#include "theodolite.h"
#include "vector.h"

namespace spex {

class Swathcal : public Processor
{
public:
    Swathcal(const gadfit::MPIVars& mpi, const SPEXSettings& settings, CKD& ckd)
      : Processor { mpi, settings, ProcLevel::swath, ckd }
    {}
    ~Swathcal() = default;

protected:
    auto processInit() -> void override;
    auto prepareBatches() -> void {}
    auto processBatch(const int i_vp) -> void override;
    auto processFinalize() -> void override;

private:
    // Administration
    std::vector<size_t> ibatch_start {};
    std::vector<size_t> nbatch_vp {};
    std::vector<size_t> nfail_vp {};
    // Raw fit results
    std::vector<double> act_fovcal {};
    std::vector<double> act_swathcal {};
    std::vector<double> alt_swathcal {};
    std::vector<bool> mask {};
    std::vector<double> mag {};
    // Intermediate output to write as detailed output
    std::vector<double> l1b_act_angles {};
    std::vector<double> l1b_alt_angles {};
    // B-spline order attempts
    size_t nattempt {};
    size_t min_fit_poorest {}; // For easy access to determine certain failure
    // Rotation quaternion
    Quaternion q_inst {}; // Instrument rotation quaternion
    std::unique_ptr<Theodolite> theo {};
    // Private function for spectrum extraction, an often-repeated action
    // during process_batch. This extraction does not work for binned images
    auto act_extract(size_t ivp, // Viewport index
                     double* image, // Detector image
                     const SharedArray<bool>& pixel_mask, // Detector pixel mask
                     double act, // ACT-angle in FOVcal perspective
                     double& strength) -> void; // Output spectrum strength (median of two spectra)
};

} // namespace spex
