/*
#include "test.h"

#include "bspline.h"
#include "linalg.h"
#include "spectra.h"

#include "tools.h"

#include <fstream>
#include <iostream>

namespace spex {

constexpr int i_fov { 100 };
constexpr bool use_mod_var { false };
// constexpr bool use_mod_var { true };

[[nodiscard]] static auto f_mod(const double p, const double m) -> double
{
    // return p - m;
    return (p - m) / (p + m);
}

static auto part1(L1A& l1a,
                  CKD& ckd,
                  std::vector<double>& radiance_fine,
                  const SPEXSettings& settings,
                  std::vector<double>& dolp_ref) -> void
{
    std::ofstream out { "/home/raul/tmp/data_ref.txt" };
    Spectra spectra {};
    {
        const auto& bin { l1a.binning_table_tmp.get() };
        std::vector<double> image_binned(bin->nBins());
        bin->binData(l1a.image, image_binned);
        // bin->unbinData(image_binned, l1a.image);
    }
    l1a.extractSpectra(i_fov, spectra);
    if (ckd.level > ProcLevel::rad) {
        for (int i {}; i < ckd.fov_dims_spec[i_fov]; ++i) {
            if (!spectra.mask[pol_minus][i] && !spectra.mask[pol_plus][i]) {
                radiance_fine[i] = spectra.signal[pol_plus][i]
                                   + spectra.signal[pol_minus][i];
            }
        }
    }
    std::vector<int> fov_start_bin(ckd.dim_fov);
    std::vector<int> fov_nmeas(ckd.dim_fov);
    for (int i_fov {}; i_fov < ckd.dim_fov; ++i_fov) {
        int ibin {};
        if (settings.l1b.wave_min) {
            while (ckd.wave_common[i_fov][ibin] < settings.l1b.wave_min) {
                ++ibin;
            }
        }
        fov_start_bin[i_fov] = ibin;
        ibin = ckd.fov_dims_spec[i_fov] - 1;
        if (settings.l1b.wave_max) {
            while (ckd.wave_common[i_fov][ibin] > settings.l1b.wave_max) {
                --ibin;
            }
        }
        fov_nmeas[i_fov] = ibin + 1 - fov_start_bin[i_fov];
    }
    const int nmeas_max { *std::max_element(fov_nmeas.cbegin(),
                          fov_nmeas.cend()) };
    std::vector<double> radiance(nmeas_max);
    const int nmeas { fov_nmeas[i_fov] };
    std::vector<bool> modulation_mask(nmeas);
    std::vector<double> modulation(nmeas);
    std::vector<double> modulation_var(nmeas);
    for (int imeas {}; imeas < nmeas; ++imeas) {
        const int i { fov_start_bin[i_fov] + imeas };
        modulation_mask[imeas] =
          spectra.mask[pol_minus][i] || spectra.mask[pol_plus][i];
        if (modulation_mask[imeas]) {
            modulation[imeas] = 0.0;
            modulation_var[imeas] = NC_FILL_DOUBLE;
        } else {
            const double m { spectra.signal[pol_minus][i] };
            const double p { spectra.signal[pol_plus][i] };
            modulation[imeas] = f_mod(p, m);
            const double ns_m { spectra.noise[pol_minus][i] };
            const double ns_p { spectra.noise[pol_plus][i] };
            const double p_m { p + m };
            const double p_m2 { p_m * p_m };
            modulation_var[imeas] = 4 * (m * m * ns_p * ns_p + p * p
                                         * ns_m * ns_m) / (p_m2 * p_m2);
            radiance[imeas] = p_m;
        }
        if (radiance[imeas] > 1e10) {
            radiance[imeas] = 0.0;
        }
    }
    if (!use_mod_var) {
        std::ranges::fill(modulation_var, 1.0);
    }
    std::vector<double> knots { settings.l1b.knots };
    for (int i {}; i < settings.l1b.order; ++i) {
        knots.insert(knots.begin(), knots.front());
        knots.push_back(knots.back());
    }
    int n_demod_states =
      2 * (static_cast<int>(knots.size()) - settings.l1b.order - 1);
    int n_demod_states_half = n_demod_states / 2;
    int demod_size_full = n_demod_states * n_demod_states;
    int demod_size_half = n_demod_states_half * n_demod_states_half;
    BSpline demod_b_spline { knots, settings.l1b.order };
    std::vector<std::unique_ptr<PackedMatrix>> demod_spline_matrix(ckd.dim_fov);
    for (int i_fov {}; i_fov < ckd.dim_fov; ++i_fov) {
        demod_spline_matrix[i_fov] = std::make_unique<PackedMatrix>(
          fov_nmeas[i_fov], n_demod_states_half);
        for (int i {}; i < fov_nmeas[i_fov]; ++i) {
            demod_b_spline.evalBasis(
              ckd.wave_common[i_fov][fov_start_bin[i_fov] + i],
              i,
              *demod_spline_matrix[i_fov]);
        }
    }
    PackedMatrix spline_matrix_q { *demod_spline_matrix[i_fov] };
    PackedMatrix spline_matrix_u { *demod_spline_matrix[i_fov] };
    for (int i {}; i < nmeas; ++i) {
        for (int k {}; k < spline_matrix_q.row_sizes[i]; ++k) {
            const int idx { spline_matrix_q.row_locations[i] + k };
            spline_matrix_q.data[idx] *=
              ckd.pol_m_q[i_fov][fov_start_bin[i_fov] + i]
              / std::sqrt(modulation_var[i])
              ;
            spline_matrix_u.data[idx] *=
              ckd.pol_m_u[i_fov][fov_start_bin[i_fov] + i]
              / std::sqrt(modulation_var[i])
              ;
        }
    }
    PackedMatrix spline_matrix_q_T { n_demod_states_half, nmeas };
    PackedMatrix spline_matrix_u_T { n_demod_states_half, nmeas };
    spline_matrix_q_T.transpose(spline_matrix_q);
    spline_matrix_u_T.transpose(spline_matrix_u);
    // BT*B is the linear inversion matrix. It is constructed
    // from four different segments.
    std::vector<double> BTB_11(demod_size_half);
    spline_matrix_q_T.multRightTransposed(spline_matrix_q_T, BTB_11);
    std::vector<double> BTB_12(demod_size_half);
    spline_matrix_q_T.multRightTransposed(spline_matrix_u_T, BTB_12);
    std::vector<double> BTB_22(demod_size_half);
    spline_matrix_u_T.multRightTransposed(spline_matrix_u_T, BTB_22);
    std::vector<double> BTB(demod_size_full);
    for (int i {}; i < n_demod_states_half; ++i) {
        for (int j {}; j < n_demod_states_half; ++j) {
            BTB[i * n_demod_states + j] =
              BTB_11[i * n_demod_states_half + j];
        }
    }
    for (int i {}; i < n_demod_states_half; ++i) {
        for (int j {}; j < n_demod_states_half; ++j) {
            BTB[i * n_demod_states + j + n_demod_states_half] =
              BTB_12[i * n_demod_states_half + j];
        }
    }
    for (int i {}; i < n_demod_states_half; ++i) {
        for (int j {}; j < n_demod_states_half; ++j) {
            BTB[(i + n_demod_states_half) * n_demod_states + j] =
              BTB_12[i * n_demod_states_half + j];
        }
    }
    for (int i {}; i < n_demod_states_half; ++i) {
        for (int j {}; j < n_demod_states_half; ++j) {
            BTB[(i + n_demod_states_half) * n_demod_states + j
                + n_demod_states_half] =
              BTB_22[i * n_demod_states_half + j];
        }
    }
    // BT*Y is the B-spline matrix multiplied with the modulation data
    std::vector<double> BTY(n_demod_states, 0.0);
    for (int i {}; i < n_demod_states_half; ++i) {
        for (int k {}; k < spline_matrix_q_T.row_sizes[i]; ++k) {
            const int packed_idx { spline_matrix_q_T.row_locations[i]
                                   + k };
            const int col_idx { spline_matrix_q_T.global_col_indices[i]
                                + k };
            BTY[i] += spline_matrix_q_T.data[packed_idx]
                      * modulation[col_idx]
                      / std::sqrt(modulation_var[col_idx])
                      ;
            BTY[i + n_demod_states_half] +=
              spline_matrix_u_T.data[packed_idx]
              * modulation[col_idx]
              / std::sqrt(modulation_var[col_idx])
              ;
        }
    }
    // Convert BTB into an upper triangular form
    std::vector<double> BTB_tri(n_demod_states * (n_demod_states + 1)
                                / 2);
    for (int j {}; j < n_demod_states; ++j) {
        for (int i {}; i <= j; ++i) {
            BTB_tri[i + (j+1)*j/2] = BTB[i * n_demod_states + j];
        }
    }
    // Do the linear inversion. x is the resulting q/u state
    // vector and Sx the corresponding covariance matrix.
    std::vector<double> Sx(demod_size_full, NC_FILL_DOUBLE);
    std::ranges::fill(Sx, 0.0);
    for (int i {}; i < n_demod_states; ++i) {
        Sx[i * n_demod_states + i] = 1.0;
    }
    std::vector<int> IPIV(n_demod_states);
    const char uplo { 'u' };
    int info {};
    dppsv_(&uplo,
           &n_demod_states,
           &n_demod_states,
           BTB_tri.data(),
           Sx.data(),
           &n_demod_states,
           &info);
    std::vector<double> x_result(n_demod_states, 0.0);
    for (int i {}; i < n_demod_states; ++i) {
        for (int k {}; k < n_demod_states; ++k) {
            x_result[i] += Sx[i * n_demod_states + k] * BTY[k];
        }
    }
    // q and u values on the fine grid
    std::vector<double> q_fine(nmeas, 0.0);
    std::vector<double> u_fine(nmeas, 0.0);
    for (int i {}; i < nmeas; ++i) {
        for (int k {}; k < demod_spline_matrix[i_fov]->row_sizes[i];
             ++k) {
            const int spl_idx {
                demod_spline_matrix[i_fov]->row_locations[i] + k
            };
            const int x_idx {
                demod_spline_matrix[i_fov]->global_col_indices[i] + k
            };
            q_fine[i] += demod_spline_matrix[i_fov]->data[spl_idx]
                         * x_result[x_idx];
            u_fine[i] += demod_spline_matrix[i_fov]->data[spl_idx]
                         * x_result[n_demod_states_half + x_idx];
        }
        const double dolp { std::sqrt(q_fine[i] * q_fine[i]
                                      + u_fine[i] * u_fine[i]) };
        dolp_ref[i] = dolp;
        out
          << ckd.wave_common[i_fov][fov_start_bin[i_fov] + i] << ' '
          << radiance[i] << ' '
          << modulation[i] << ' '
          << q_fine[i] << ' '
          << dolp << ' '
          << '\n';
    }
}

static auto binSpectral(
  const CKD& ckd, const L1A& l1a, std::vector<double>& data) -> void
{
    const auto& bin { l1a.binning_table_tmp.get() };
    std::vector<double> data_red(2048, 0.0);
    for (int i {}; i < 2048; ++i) {
        data_red[bin->spectral_bin_indices[i_fov][i]] += data[i];
    }
    for (int i {}; i < ckd.fov_dims_spec[i_fov]; ++i) {
        data_red[i] /= bin->spectral_count_table[i_fov][i];
    }
    data_red.resize(ckd.fov_dims_spec[i_fov]);
    data = std::move(data_red);
}

auto run(L1A& l1a, CKD& ckd, const SPEXSettings& settings,
         const gadfit::MPIVars& mpi, std::vector<double>& radiance_fine) -> void
{
    // return;
    std::vector<double> dolp_ref(2048);
    part1(l1a, ckd, radiance_fine, settings, dolp_ref);
    Spectra spectra_full {};
    l1a.extractSpectra(i_fov, spectra_full);
    std::vector<double> modulation_full(2048);
    std::vector<double> modulation_var_full(2048);
    std::vector<double> p_full(2048);
    std::vector<double> m_full(2048);
    std::vector<bool> modulation_mask_full(2048);
    for (int imeas {}; imeas < 2048; ++imeas) {
        const int i { imeas };
        modulation_mask_full[imeas] =
          spectra_full.mask[pol_minus][i] || spectra_full.mask[pol_plus][i];
        if (modulation_mask_full[imeas]) {
            modulation_full[imeas] = 0.0;
            modulation_var_full[imeas] = NC_FILL_DOUBLE;
        } else {
            double m { spectra_full.signal[pol_minus][i] };
            double p { spectra_full.signal[pol_plus][i] };
            p_full[i] = p;
            m_full[i] = m;
            modulation_full[imeas] = f_mod(p, m);
            const double ns_m { spectra_full.noise[pol_minus][i] };
            const double ns_p { spectra_full.noise[pol_plus][i] };
            const double p_m { p + m };
            const double p_m2 { p_m * p_m };
            modulation_var_full[imeas] = 4 * (m * m * ns_p * ns_p + p * p
                                              * ns_m * ns_m) / (p_m2 * p_m2);
        }
    }
    if (!use_mod_var) {
        std::ranges::fill(modulation_var_full, 1.0);
    }
    std::vector<double> wave_common_save(2048);
    for (int i {}; i < 2048; ++i) {
        wave_common_save[i] = ckd.wave_common[i_fov][i];
    }
    const auto& bin { l1a.binning_table_tmp.get() };
    std::vector<double> image_binned(bin->nBins());
    bin->binData(l1a.image, image_binned);
    SharedArray<double> new_image { true, mpi, bin->nBins() };
    for (int i {}; i < static_cast<int>(new_image.size()); ++i) {
        new_image[i] = image_binned[i];
    }
    l1a.image = std::move(new_image);

    std::ofstream out { "/home/raul/tmp/data.txt" };

    std::vector<double> knots { settings.l1b.knots };
    for (int i {}; i < settings.l1b.order; ++i) {
        knots.insert(knots.begin(), knots.front());
        knots.push_back(knots.back());
    }
    // We need to subtract (order+1) because the endpoint (the
    // last i+1) value is not included.
    int n_demod_states =
      2 * (static_cast<int>(knots.size()) - settings.l1b.order - 1);
    int n_demod_states_half = n_demod_states / 2;
    int demod_size_full = n_demod_states * n_demod_states;
    int demod_size_half = n_demod_states_half * n_demod_states_half;
    BSpline demod_b_spline { knots, settings.l1b.order };

    bin->transformFOVPixels(ckd);
    bin->generateSpectralBinningTable(ckd);
    for (int i_pol {}; i_pol < dims::pol; ++i_pol) {
        // bin->binSpectral(ckd, ckd.fov_ipix1.at(i_pol));
        // bin->binSpectral(ckd, ckd.fov_ipix2.at(i_pol));
        // bin->binSpectral(ckd, ckd.fov_weight1.at(i_pol));
        bin->binSpectral(ckd, ckd.wave_full.at(i_pol));
    }
    // FULL START
    std::vector<double> demod_spline_matrix_full(2048 * n_demod_states_half);
    std::vector<double> spline_matrix_q_full(2048 * n_demod_states_half);
    std::vector<double> spline_matrix_u_full(2048 * n_demod_states_half);
    for (int i {}; i < 2048; ++i) {
        for (int j {}; j < n_demod_states_half; ++j) {
            demod_b_spline.evalBasis(
              ckd.wave_common[i_fov][i],
              &demod_spline_matrix_full[i * n_demod_states_half]);
            spline_matrix_q_full[i * n_demod_states_half + j] =
              demod_spline_matrix_full[i * n_demod_states_half + j]
              * ckd.pol_m_q[i_fov][i]
              // / std::sqrt(modulation_var[i])
              ;
            spline_matrix_u_full[i * n_demod_states_half + j] =
              demod_spline_matrix_full[i * n_demod_states_half + j]
              * ckd.pol_m_u[i_fov][i];
              // / std::sqrt(modulation_var[i])
        }
    }
    std::vector<double> demod_spline_matrix_red(2048 * n_demod_states_half,0.0);
    std::vector<double> spline_matrix_q_red(2048 * n_demod_states_half, 0.0);
    std::vector<double> spline_matrix_u_red(2048 * n_demod_states_half, 0.0);
    for (int j {}; j < n_demod_states_half; ++j) {
        for (int i {}; i < 2048; ++i) {
            const uint32_t idx {
                bin->spectral_bin_indices[i_fov][i]
            };
            demod_spline_matrix_red[idx * n_demod_states_half + j] +=
              demod_spline_matrix_full[i * n_demod_states_half + j];
            spline_matrix_q_red[idx * n_demod_states_half + j] +=
              spline_matrix_q_full[i * n_demod_states_half + j];
            spline_matrix_u_red[idx * n_demod_states_half + j] +=
              spline_matrix_u_full[i * n_demod_states_half + j];
        }
        for (int i {}; i < ckd.fov_dims_spec[i_fov]; ++i) {
            const uint16_t count {
                bin->spectral_count_table[i_fov][i]
            };
            demod_spline_matrix_red[i * n_demod_states_half + j] /= count;
            spline_matrix_q_red[i * n_demod_states_half + j] /= count;
            spline_matrix_u_red[i * n_demod_states_half + j] /= count;
        }
    }
    demod_spline_matrix_red.resize(ckd.fov_dims_spec[i_fov]
                                   * n_demod_states_half);
    spline_matrix_q_red.resize(ckd.fov_dims_spec[i_fov] * n_demod_states_half);
    spline_matrix_u_red.resize(ckd.fov_dims_spec[i_fov] * n_demod_states_half);
    binSpectral(ckd, l1a, modulation_full);
    // FULL END
    bin->binSpectral(ckd, ckd.wave_common);
    std::vector<int> fov_start_bin(ckd.dim_fov);
    std::vector<int> fov_nmeas(ckd.dim_fov);
    for (int i_fov {}; i_fov < ckd.dim_fov; ++i_fov) {
        int ibin {};
        if (settings.l1b.wave_min) {
            while (ckd.wave_common[i_fov][ibin] < settings.l1b.wave_min) {
                ++ibin;
            }
        }
        fov_start_bin[i_fov] = ibin;
        ibin = ckd.fov_dims_spec[i_fov] - 1;
        if (settings.l1b.wave_max) {
            while (ckd.wave_common[i_fov][ibin] > settings.l1b.wave_max) {
                --ibin;
            }
        }
        fov_nmeas[i_fov] = ibin + 1 - fov_start_bin[i_fov];
    }
    for (int i_pol {}; i_pol < 2; ++i_pol) {
        bin->binSpectral(ckd, ckd.rad_spectra[i_pol]);
    }
    bin->binSpectral(ckd, ckd.pol_m_q);
    bin->binSpectral(ckd, ckd.pol_m_u);
    bin->binSpectral(ckd, ckd.pol_m_t);
    std::vector<std::vector<double>> demod_spline_matrix0(ckd.dim_fov);
    for (int i_fov {}; i_fov < ckd.dim_fov; ++i_fov) {
        demod_spline_matrix0[i_fov].resize(fov_nmeas[i_fov]
                                           * n_demod_states_half);
        for (int i {}; i < fov_nmeas[i_fov]; ++i) {
            demod_b_spline.evalBasis(
              ckd.wave_common[i_fov][fov_start_bin[i_fov] + i],
              &demod_spline_matrix0[i_fov][i * n_demod_states_half]);
        }
    }

    const int nmeas_max { *std::max_element(fov_nmeas.cbegin(),
                          fov_nmeas.cend()) };
    Spectra spectra {};
    l1a.extractSpectra(i_fov, spectra);
    const int nmeas { fov_nmeas[i_fov] };
    std::vector<bool> modulation_mask(nmeas_max);
    std::vector<double> radiance(nmeas_max);
    std::vector<double> modulation(nmeas_max);
    std::vector<double> modulation_var(nmeas_max);
    for (int imeas {}; imeas < nmeas; ++imeas) {
        const int i { fov_start_bin[i_fov] + imeas };
        modulation_mask[imeas] =
          spectra.mask[pol_minus][i] || spectra.mask[pol_plus][i];
        if (modulation_mask[imeas]) {
            modulation[imeas] = 0.0;
            modulation_var[imeas] = NC_FILL_DOUBLE;
        } else {
            const double m { spectra.signal[pol_minus][i] };
            const double p { spectra.signal[pol_plus][i] };
            modulation[imeas] = f_mod(p, m);
            const double ns_m { spectra.noise[pol_minus][i] };
            const double ns_p { spectra.noise[pol_plus][i] };
            const double p_m { p + m };
            radiance[imeas] = p_m;
            const double p_m2 { p_m * p_m };
            modulation_var[imeas] = 4 * (m * m * ns_p * ns_p + p * p
                                         * ns_m * ns_m) / (p_m2 * p_m2);
        }
        if (radiance[imeas] > 1e10) {
            radiance[imeas] = 0.0;
        }
    }
    if (!use_mod_var) {
        std::ranges::fill(modulation_var, 1.0);
    }
    std::vector<double> dolp_full(2048);
    binSpectral(ckd, l1a, dolp_ref);
    for (int iter {}; iter < 2; ++iter) {
        if (iter == 1) {
            modulation = modulation_full;
            modulation_var = modulation_var_full;
        }
        std::vector<double> spline_matrix_q0(nmeas * n_demod_states_half);
        std::vector<double> spline_matrix_u0(nmeas * n_demod_states_half);
        // for (int i {}; i < nmeas; ++i) {
        //     for (int j {}; j < n_demod_states_half; ++j) {
        //         spline_matrix_q0[i * n_demod_states_half + j] =
        //           demod_spline_matrix0[i_fov][i * n_demod_states_half + j]
        //           * ckd.pol_m_q[i_fov][fov_start_bin[i_fov] + i]
        //           // / std::sqrt(modulation_var[i])
        //           ;
        //         spline_matrix_u0[i * n_demod_states_half + j] =
        //           demod_spline_matrix0[i_fov][i * n_demod_states_half + j]
        //           * ckd.pol_m_u[i_fov][fov_start_bin[i_fov] + i]
        //           // / std::sqrt(modulation_var[i])
        //           ;
        //     }
        // }
        demod_spline_matrix0[i_fov] = demod_spline_matrix_red;
        spline_matrix_q0 = spline_matrix_q_red;
        // writeBin("/home/raul/tmp/image.bin", demod_spline_matrix0[i_fov]);
        spline_matrix_u0 = spline_matrix_u_red;
        // PackedMatrix spline_matrix_q_T { n_demod_states_half, nmeas };
        // PackedMatrix spline_matrix_u_T { n_demod_states_half, nmeas };
        // spline_matrix_q_T.transpose(spline_matrix_q);
        // spline_matrix_u_T.transpose(spline_matrix_u);
        // BT*B is the linear inversion matrix. It is constructed
        // from four different segments.
        std::vector<double> BTB_11(demod_size_half, 0.0);
        // spline_matrix_q_T.multRightTransposed(spline_matrix_q_T, BTB_11);
        std::vector<double> BTB_12(demod_size_half, 0.0);
        // spline_matrix_q_T.multRightTransposed(spline_matrix_u_T, BTB_12);
        std::vector<double> BTB_22(demod_size_half, 0.0);
        // spline_matrix_u_T.multRightTransposed(spline_matrix_u_T, BTB_22);
        for (int i {}; i < n_demod_states_half; ++i) {
            for (int j {}; j < n_demod_states_half; ++j) {
                for (int k {}; k < nmeas; ++k) {
                    BTB_11[i * n_demod_states_half + j] +=
                      spline_matrix_q0[k * n_demod_states_half + i]
                      * spline_matrix_q0[k * n_demod_states_half + j];
                    BTB_12[i * n_demod_states_half + j] +=
                      spline_matrix_q0[k * n_demod_states_half + i]
                      * spline_matrix_u0[k * n_demod_states_half + j];
                    BTB_22[i * n_demod_states_half + j] +=
                      spline_matrix_u0[k * n_demod_states_half + i]
                      * spline_matrix_u0[k * n_demod_states_half + j];
                }
            }
        }
        std::vector<double> BTB(demod_size_full);
        for (int i {}; i < n_demod_states_half; ++i) {
            for (int j {}; j < n_demod_states_half; ++j) {
                BTB[i * n_demod_states + j] =
                  BTB_11[i * n_demod_states_half + j];
            }
        }
        for (int i {}; i < n_demod_states_half; ++i) {
            for (int j {}; j < n_demod_states_half; ++j) {
                BTB[i * n_demod_states + j + n_demod_states_half] =
                  BTB_12[i * n_demod_states_half + j];
            }
        }
        for (int i {}; i < n_demod_states_half; ++i) {
            for (int j {}; j < n_demod_states_half; ++j) {
                BTB[(i + n_demod_states_half) * n_demod_states + j] =
                  BTB_12[i * n_demod_states_half + j];
            }
        }
        for (int i {}; i < n_demod_states_half; ++i) {
            for (int j {}; j < n_demod_states_half; ++j) {
                BTB[(i + n_demod_states_half) * n_demod_states + j
                    + n_demod_states_half] =
                  BTB_22[i * n_demod_states_half + j];
            }
        }
        // BT*Y is the B-spline matrix multiplied with the modulation data
        std::vector<double> BTY(n_demod_states, 0.0);
        for (int i {}; i < n_demod_states_half; ++i) {
            for (int j {}; j < nmeas; ++j) {
                BTY[i] += spline_matrix_q0[j * n_demod_states_half + i]
                          * modulation[j]
                          / std::sqrt(modulation_var[j])
                          ;
                BTY[i + n_demod_states_half]
                  += spline_matrix_u0[j * n_demod_states_half + i]
                  * modulation[j]
                  / std::sqrt(modulation_var[j])
                  ;
            }
        }
        // Convert BTB into an upper triangular form
        std::vector<double> BTB_tri(n_demod_states * (n_demod_states + 1)
                                    / 2);
        for (int j {}; j < n_demod_states; ++j) {
            for (int i {}; i <= j; ++i) {
                BTB_tri[i + (j+1)*j/2] = BTB[i * n_demod_states + j];
            }
        }
        // Do the linear inversion. x is the resulting q/u state
        // vector and Sx the corresponding covariance matrix.
        std::vector<double> Sx(demod_size_full, NC_FILL_DOUBLE);
        std::ranges::fill(Sx, 0.0);
        for (int i {}; i < n_demod_states; ++i) {
            Sx[i * n_demod_states + i] = 1.0;
        }
        std::vector<int> IPIV(n_demod_states);
        const char uplo { 'u' };
        int info {};
        dppsv_(&uplo,
               &n_demod_states,
               &n_demod_states,
               BTB_tri.data(),
               Sx.data(),
               &n_demod_states,
               &info);
        std::vector<double> x_result(n_demod_states, 0.0);
        for (int i {}; i < n_demod_states; ++i) {
            for (int k {}; k < n_demod_states; ++k) {
                x_result[i] += Sx[i * n_demod_states + k] * BTY[k];
            }
        }
        // q and u values on the fine grid
        std::vector<double> q_fine(nmeas, 0.0);
        std::vector<double> u_fine(nmeas, 0.0);
        for (int i {}; i < nmeas; ++i) {
            for (int j {}; j < n_demod_states_half; ++j) {
                q_fine[i] +=
                  demod_spline_matrix0[i_fov][i * n_demod_states_half +j]
                  * x_result[j];
                u_fine[i]
                  += demod_spline_matrix0[i_fov][i * n_demod_states_half +j]
                             * x_result[n_demod_states_half + j];
            }
            const double dolp { std::sqrt(q_fine[i] * q_fine[i]
                                          + u_fine[i] * u_fine[i]) };
            if (iter == 0) {
                out
                  << ckd.wave_common[i_fov][fov_start_bin[i_fov] + i] << ' '
                  << radiance[i] << ' '
                  << modulation[i] << ' '
                  << modulation_full[i] << ' '
                  << q_fine[i] << ' '
                  << dolp << ' '
                  << dolp_ref[i] << ' '
                  << '\n';
            } else {
                dolp_full[i] = dolp;
            }
        }
    }
    out << '\n' << '\n';
    for (int i {}; i < nmeas; ++i) {
        out
          << ckd.wave_common[i_fov][fov_start_bin[i_fov] + i] << ' '
          << dolp_full[i] << ' '
          << dolp_ref[i] << ' '
          << '\n';
    }
    out.close();
    MPI_Finalize();exit(0);
}

} // namespace spex
*/
