#pragma once
// MPI_Finalize();exit(0);
// std::ofstream out { "/home/raul/tmp/data.txt" };

#include <math.h>

#include <fstream>
#include <iomanip>
template <typename T>
static auto writeBin(const std::string& filename, const T& M, int n_rows = 0)
  -> void
{
    if (n_rows == 0) {
        n_rows = static_cast<int>(std::round(std::sqrt(M.size())));
    }
    try {
        std::ofstream wf(filename, std::ios::binary);
        wf.exceptions(std::ofstream::failbit);
        const double d_rows { static_cast<double>(n_rows) };
        wf.write((char*)(&d_rows), sizeof(double));
        wf.write((char*)&M[0], M.size() * sizeof(double));
    } catch (const std::ofstream::failure& e) {
        std::cerr << "Cannot create file: " << filename << '\n';
        throw;
    }
}

template <typename T>
static auto writeBin(const T& M, int n_rows = 0) -> void
{
    writeBin("/home/raul/tmp/image.bin", M, n_rows);
}

template <typename T>
auto readBin(const std::string& filename, T& M) -> void
{
    try {
        std::ifstream wf(filename, std::ios::binary);
        wf.exceptions(std::ifstream::failbit);
        double d_rows {};
        wf.read((char*)(&d_rows), sizeof(double));
        wf.read((char*)M.data(), M.size() * sizeof(double));
    } catch (const std::ofstream::failure& e) {
        std::cerr << "Could not open file: " << filename << '\n';
        throw;
    }
}

template <typename T>
auto print(const T& A,
           const int n_rows,
           const int n_cols,
           const int precision,
           int row_dim = 0,
           const int row_start = 0,
           const int col_start = 0,
           const bool transpose = false) -> void
{
    if (row_dim == 0) {
        row_dim = static_cast<int>(std::round(std::sqrt(A.size())));
    }
    const int col_dim { static_cast<int>(A.size()) / row_dim };
    std::cout.precision(precision);
    if (transpose) {
        for (int j { col_start }; j < col_start + n_cols; ++j) {
            for (int i { row_start }; i < row_start + n_rows; ++i) {
                std::cout << std::setw(precision + 7) << A[i * col_dim + j];
            }
            std::cout << '\n';
        }
    } else {
        for (int i { row_start }; i < row_start + n_rows; ++i) {
            for (int j { col_start }; j < col_start + n_cols; ++j) {
                std::cout << std::setw(precision + 7) << A[i * col_dim + j];
            }
            std::cout << '\n';
        }
    }
}
