// Generation of FOV CKD

#include "testing.h"

#include <spexone_cal/ckd.h>
#include <spexone_cal/driver.h>
#include <spexone_cal/spex_settings.h>

TEST_CASE("FOV")
{
    // Main run
    spex::SPEXSettings settings { CMAKE_CURRENT_BINARY_DIR "/fov_gen.yaml" };
    settings.init();
    MPI_Comm mpi_comm { MPI_COMM_WORLD };
    spex::driver(mpi_comm, settings);

    // Read the generated CKD
    settings.main.ckd_file_in = settings.main.ckd_file_out.c_str();
    const gadfit::MPIVars mpi { mpi_comm };
    spex::CKD ckd { mpi, spex::ProcLevel::n_levels, settings };

    // Read the reference CKD
    settings.main.ckd_file_in = TEST_DATA_DIR "/fov_gen/ckd_ref.nc";
    spex::CKD ckd_ref { mpi, spex::ProcLevel::n_levels, settings };

    // Must wait because the root process reads data into shared
    // arrays that are accessible to all other processes.
#ifdef USE_MPI
    MPI_Barrier(mpi.comm);
#endif

    REQUIRE(std::ranges::equal(ckd.pixel_mask, ckd_ref.pixel_mask));
    REQUIRE(std::ranges::equal(ckd.fov.n_act_vp, ckd_ref.fov.n_act_vp));
    REQUIRE(std::ranges::equal(ckd.fov.start_vp, ckd_ref.fov.start_vp));
    for (int i_pol {}; i_pol < spex::dims::pol; ++i_pol) {
        for (int i_act {}; i_act < ckd.fov.n_act; ++i_act) {
            REQUIRE(mae(ckd.fov.spat_indices.at(i_pol).at(i_act),
                        ckd_ref.fov.spat_indices.at(i_pol).at(i_act))
                    < 1e-8);
            REQUIRE(std::ranges::equal(ckd.fov.pix1.at(i_pol).at(i_act),
                                       ckd_ref.fov.pix1.at(i_pol).at(i_act)));
            REQUIRE(std::ranges::equal(ckd.fov.pix2.at(i_pol).at(i_act),
                                       ckd_ref.fov.pix2.at(i_pol).at(i_act)));
            REQUIRE(mae(ckd.fov.weight.at(i_pol).at(i_act),
                        ckd_ref.fov.weight.at(i_pol).at(i_act))
                    < 1e-9);
            REQUIRE(mae(ckd.fov.peak_pos.at(i_pol).at(i_act),
                        ckd_ref.fov.peak_pos.at(i_pol).at(i_act))
                    < 1e-12);
            REQUIRE(mae(ckd.fov.peak_pos_b.at(i_pol).at(i_act),
                        ckd_ref.fov.peak_pos_b.at(i_pol).at(i_act))
                    < 1e-11);
        }
    }
}
