// Generation of noise CKD

#include "testing.h"

#include <spexone_cal/ckd.h>
#include <spexone_cal/driver.h>
#include <spexone_cal/spex_settings.h>

TEST_CASE("Noise")
{
    // Main run
    spex::SPEXSettings settings { CMAKE_CURRENT_BINARY_DIR "/noise_gen.yaml" };
    settings.init();
    MPI_Comm mpi_comm { MPI_COMM_WORLD };
    spex::driver(mpi_comm, settings);

    // Read the generated CKD
    settings.main.ckd_file_in = settings.main.ckd_file_out.c_str();
    const gadfit::MPIVars mpi { mpi_comm };
    spex::CKD ckd { mpi, spex::ProcLevel::n_levels, settings };

    // Read the reference CKD
    settings.main.ckd_file_in = TEST_DATA_DIR "/noise_gen/ckd_ref.nc";
    spex::CKD ckd_ref { mpi, spex::ProcLevel::n_levels, settings };

    // Must wait because the root process reads data into shared
    // arrays that are accessible to all other processes.
#ifdef USE_MPI
    MPI_Barrier(mpi.comm);
#endif

    REQUIRE(std::ranges::equal(ckd.pixel_mask, ckd_ref.pixel_mask));
    REQUIRE(mae(ckd.noise.g, ckd_ref.noise.g) < 1e-6);
    REQUIRE(mae(ckd.noise.n, ckd_ref.noise.n) < 1e-6);
    REQUIRE(mae(ckd.noise.debug.std_dev_g, ckd_ref.noise.debug.std_dev_g)
            < 1e-6);
    REQUIRE(mae(ckd.noise.debug.std_dev_n, ckd_ref.noise.debug.std_dev_n)
            < 1e-3);
}
