// Generation of polarimetric CKD

#include "testing.h"

#include <spexone_cal/ckd.h>
#include <spexone_cal/driver.h>
#include <spexone_cal/spex_settings.h>

TEST_CASE("Polarimetric")
{
    // Main run
    spex::SPEXSettings settings { CMAKE_CURRENT_BINARY_DIR "/pol_gen.yaml" };
    settings.init();
    MPI_Comm mpi_comm { MPI_COMM_WORLD };
    spex::driver(mpi_comm, settings);

    // Read the generated CKD
    settings.main.ckd_file_in = settings.main.ckd_file_out.c_str();
    const gadfit::MPIVars mpi { mpi_comm };
    spex::CKD ckd { mpi, spex::ProcLevel::n_levels, settings };

    // Read the reference CKD
    settings.main.ckd_file_in = TEST_DATA_DIR "/pol_gen/ckd_ref.nc";
    spex::CKD ckd_ref { mpi, spex::ProcLevel::n_levels, settings };

    // Must wait because the root process reads data into shared
    // arrays that are accessible to all other processes.
#ifdef USE_MPI
    MPI_Barrier(mpi.comm);
#endif

    REQUIRE(std::ranges::equal(ckd.pixel_mask, ckd_ref.pixel_mask));
    for (int i {}; i < ckd.fov.n_act; ++i) {
        REQUIRE(mae(ckd.pol.m_q.at(i), ckd_ref.pol.m_q.at(i)) < 1e-3);
        REQUIRE(mae(ckd.pol.m_u.at(i), ckd_ref.pol.m_u.at(i)) < 1e-4);
        REQUIRE(mae(ckd.pol.m_t.at(i), ckd_ref.pol.m_t.at(i)) < 1e-4);
        REQUIRE(mae(ckd.pol.debug.eps.at(i), ckd_ref.pol.debug.eps.at(i))
                < 1e-6);
        REQUIRE(mae(ckd.pol.debug.delta.at(i), ckd_ref.pol.debug.delta.at(i))
                < 1e-8);
    }
}
